#include <Arduino.h>
#include <nRF24L01.h>
#include <RF24.h>
#include "config.h"
#include "radio.h"
#include "clockMsg.h"


static RF24 rf24(PIN_NRF24_CE, PIN_NRF24_CSN); // 10, 8
static byte addresses[][6] = {NETWORK_ADDRESS_MASTER_SEND, NETWORK_ADDRESS_MASTER_RECEIVE};

void Radio::begin(void) {
  display->addLogMessage("Start RF24 radio");
  pinMode(PIN_NRF24_CSN, OUTPUT);
  pinMode(PIN_NRF24_CE, OUTPUT);

  rf24.begin();
  if (rf24.isChipConnected()) { display->addLogMessage("*** RF chip found"); }
  else { display->addLogMessage("*** ERROR: RF chip not found!"); }
  rf24.setChannel(1);
  rf24.setPALevel(RF24_PA_MAX);
  rf24.setDataRate(RF24_2MBPS);
  rf24.setAutoAck(0);
  //radio.setRetries(2,15); // Optionally, increase the delay between retries & # of retries
  rf24.setCRCLength(RF24_CRC_8);
  rf24.openWritingPipe(addresses[0]);
  rf24.openReadingPipe(1, addresses[1]);
  rf24.startListening();
  Serial.println("*** Check");
  //rf24.printDetails();
  rf24.powerUp();
  // @TODO: real random seed!
  //randomSeed(analogRead(0));
  //randomSeed(22);
  Serial.print("*** RF payload size="); Serial.print(rf24.getPayloadSize()); Serial.println(" bytes");
  if (rf24.testCarrier() || rf24.testRPD()) { display->addLogMessage("*** Carrier/RPD seen on radio"); }
  if (rf24.failureDetected) { display->addLogMessage("*** Radio error detected!"); }


#if 0
  // Open a writing and reading pipe on each radio, with opposite addresses
  if (isMaster){
    rf24.openWritingPipe(addresses[1]);
    rf24.openReadingPipe(1,addresses[0]);
  } else {
    rf24.openWritingPipe(addresses[0]);
    rf24.openReadingPipe(1,addresses[1]);
  }

#endif

  // Start the radio listening for data
  rf24.startListening();
}

static bool inTX=1, inRX=0, role=inRX;

static void switchToSenderRole(RF24 radio)
{
  Serial.println("*** master: CHANGING TO TRANSMIT ROLE");
  radio.openWritingPipe(addresses[1]);
  radio.openReadingPipe(1,addresses[0]);
  radio.stopListening();
  role = inTX;                  // Become the primary transmitter (ping out)
}

static void switchToReceiverRole(RF24 radio)
{
  Serial.println("*** master: CHANGING TO RECEIVER ROLE");
  radio.openWritingPipe(addresses[0]);
  radio.openReadingPipe(1,addresses[1]);
  radio.startListening();
  role = inRX;                // Become the primary receiver (pong back)
}

static int sendFailedCounter = 0;
static unsigned long stopTime = 0, pauseTime = 0;

void Radio::broadcastRequestRegistration(const char *clockName) {
  char sendBuffer[32];

  memset(sendBuffer, 0, 32);
  sendBuffer[0] = msgType_ReqReg;
  strncpy(sendBuffer+1, clockName, MAX_CLOCK_NAME_LEN);
  int msgLength = MAX_CLOCK_NAME_LEN + 1;

  switchToSenderRole(rf24);
  if (!rf24.writeFast(sendBuffer, msgLength, true /*multicast*/)) {
      sendFailedCounter++;
      Serial.print("*** ERROR: failed to send ReqRegistration msg for "); Serial.println(clockName);
  }
  //This is only required when NO ACK ( enableAutoAck(0) ) payloads are used
  if (millis() - pauseTime > 3) {
    pauseTime = millis();
    rf24.txStandBy();          // Need to drop out of TX mode every 4ms if sending a steady stream of multicast data
    //delayMicroseconds(130);     // This gives the PLL time to sync back up
  }
  stopTime = millis();
                                      //This should be called to wait for completion and put the radio in standby mode after transmission, returns 0 if data still in FIFO (timed out), 1 if success
  if (!rf24.txStandBy()) { sendFailedCounter += 3; } //Standby, block only until FIFO empty or auto-retry timeout. Flush TX FIFO if failed
  //radio.txStandBy(1000);              //Standby, using extended timeout period of 1 second

  Serial.print("*** send finished, fail-counter="); Serial.println(sendFailedCounter);
}

void Radio::broadcastClock(int day, int hour, int minute, int second, int msPerModelSecond) {
  struct clockMsg_s clockMsg;

  clockMsg.msgType = msgType_Clock;
  clockMsg.day = day;
  clockMsg.hour = hour;
  clockMsg.minute = minute;
  clockMsg.second = second;
  clockMsg.speed = (msPerModelSecond >> 2) & 0xff; // divide by 4
  switchToSenderRole(rf24);
  if (!rf24.writeFast(&clockMsg, sizeof(clockMsg), true /*multicast*/)) {
      sendFailedCounter++;
      Serial.print("*** ERROR: failed to send clock msg for "); Serial.print(hour); Serial.print(":"); Serial.print(minute); Serial.print(":"); Serial.println(second);
  }
  //This is only required when NO ACK ( enableAutoAck(0) ) payloads are used
  if (millis() - pauseTime > 3) {
    pauseTime = millis();
    rf24.txStandBy();          // Need to drop out of TX mode every 4ms if sending a steady stream of multicast data
    //delayMicroseconds(130);     // This gives the PLL time to sync back up
  }
  stopTime = millis();
                                      //This should be called to wait for completion and put the radio in standby mode after transmission, returns 0 if data still in FIFO (timed out), 1 if success
  if (!rf24.txStandBy()) { sendFailedCounter += 3; } //Standby, block only until FIFO empty or auto-retry timeout. Flush TX FIFO if failed
  //radio.txStandBy(1000);              //Standby, using extended timeout period of 1 second

  Serial.print("*** send finished, fail-counter="); Serial.println(sendFailedCounter);
}

void Radio::loop(void) {
  // put your main code here, to run repeatedly:
  /****************** Ping Out Role ***************************/
  if (isMaster)  {
    switchToReceiverRole(rf24);
    if (rf24.available()) {
      char buffer[32];
      rf24.read(buffer, 32);

      // Spew it
      Serial.print(F("Received new message, type="));
      Serial.print(buffer[0]);
      Serial.print(F(", bytes 1="));
      Serial.print(buffer[1]);
      Serial.print(", 2=");
      Serial.print(buffer[2]);
      Serial.print(", 3=");
      Serial.println(buffer[3]);
    }
  }



    #if 0

  /****************** Pong Back Role ***************************/

    if (!isMaster)
    {
      unsigned long got_time;

      if( radio.available()){
                                                                      // Variable for the received timestamp
        while (radio.available()) {                                   // While there is data ready
          radio.read( &got_time, sizeof(unsigned long) );             // Get the payload
        }

        radio.stopListening();                                        // First, stop listening so we can talk
        radio.write( &got_time, sizeof(unsigned long) );              // Send the final one back.
        radio.startListening();                                       // Now, resume listening so we catch the next packets.
        Serial.print(F("Sent response "));
        Serial.println(got_time);
     }
   }
   #endif
}
